"""
Helper to make api queries.

Only does GET requests for now.

This is a simple wrapper; apart from the initial api instance, this only passes
dicts and json around.

TODO move out here! or use someone else's Python SDK. I'm lazy.

docs: https://github.com/WP-API/WP-API/blob/master/docs/guides/getting-started.md
"""
import logging

import requests
import six

logger = logging.getLogger(__name__)


class WPApiException(Exception):
    pass


class UserCannotReadException(WPApiException):
    """
    Http 401 for when api does not have the auth to read a resource.
    """
    _code = 'json_user_cannot_read'
    message = 'Sorry, you cannot read this post.'


class InvalidPostIDException(WPApiException):
    """
    Http 404 for when a post does not exist.
    """
    _code = 'json_post_invalid_id'
    message = 'Invalid post ID.'


class WPApi(object):
    _total = None
    _total_page = None

    def __init__(self, url):
        """
        Initiate connection to a site.

        Site must have the `json-rest-api` plugin installed.
        """
        self.base_url = url.rstrip('/')

    def interpret_header(self, response, header):
        """If the header value is an integer, cast it."""
        out = response.headers.get(header)
        if out is not None:
            return int(out)
        return out

    def get(self, *args):
        url = u'{}/wp-json/'.format(self.base_url) + u'/'.join(args)
        response = requests.get(url)
        if not response.ok:
            if response.status_code == 401:
                raise UserCannotReadException()
            raise WPApiException()
        self.response = response  # store last response as an attribute
        self._total = self.interpret_header(response, 'X-WP-Total')
        self._total_pages = self.interpret_header(response, 'X-WP-TotalPages')
        logger.info(u'get {}'.format(url), extra={
            'headers': response.headers,
            'status_code': response.status_code,
            'response_text': response.text,
            'elapsed': response.elapsed,
        })
        return response.json()

    def index(self):
        """
        Get information about the site.

        https://raw.githubusercontent.com/WP-API/WP-API/master/docs/schema.json
        """
        return self.get()

    def posts(self, wp_id=None):
        """
        Get a list of posts.

        To get
        TODO handle pagination via the headers

        docs: https://github.com/WP-API/WP-API/blob/master/docs/guides/working-with-posts.md
        """
        if wp_id is None:
            # get a collection
            return self.get('posts')
        else:
            return self.get('posts', six.text_type(wp_id))

    def users(self):
        """
        Get a list of users.
        """
        return self.get('users')



import factory

from django.utils.timezone import now

from .models import (
    WPSite,
    WPUser,
    WPTag,
    WPCategory,
    WPPost,
)


__all__ = ['WPSiteFactory', 'WPPostFactory']


class WPSiteFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = WPSite

    url = factory.Sequence(lambda i: 'http://{}.site'.format(i))


class WPUserFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = WPUser

    wp = factory.SubFactory(WPSiteFactory)
    id = factory.Sequence(int)
    synced_at = factory.LazyAttribute(lambda __: now())
    url = factory.Sequence(lambda i: 'http://{}.user'.format(i))


class WPTagFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = WPTag

    wp = factory.SubFactory(WPSiteFactory)
    id = factory.Sequence(int)
    synced_at = factory.LazyAttribute(lambda __: now())
    link = factory.Sequence(lambda i: 'http://{}.link'.format(i))


class WPCategoryFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = WPCategory

    wp = factory.SubFactory(WPSiteFactory)
    id = factory.Sequence(int)
    synced_at = factory.LazyAttribute(lambda __: now())
    link = factory.Sequence(lambda i: 'http://{}.link'.format(i))


class WPPostFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = WPPost

    wp = factory.SubFactory(WPSiteFactory)
    id = factory.Sequence(int)
    synced_at = factory.LazyAttribute(lambda __: now())
    date = factory.LazyAttribute(lambda __: now())
    modified = factory.LazyAttribute(lambda __: now())
    link = factory.Sequence(lambda i: 'http://{}.post'.format(i))



"""

field names are taken to match the json output (except always lowercase). This
includes `id`. For the true primary key, use the `pk` attribute. But hey, you
were already doing that, right? This makes it so that matching fields between
the Django models and the JSON resource always match.

`max_length`s of 255 means I just picked an arbitrary high number and does not
reflect some real schema limit.
"""
from django.core.urlresolvers import reverse
from django.db import models

from . import managers
from .api import WPApi
from .dj_exten_jsonfield import JSONField


###################
# ABSTRACT MODELS #
###################

class WPObjectModel(models.Model):
    wp = models.ForeignKey('WPSite')
    # not the primary key (see `dj_id`) so fields match JSON response
    id = models.PositiveIntegerField()

    # bookkeepping
    dj_id = models.AutoField(primary_key=True)
    synced_at = models.DateTimeField()  # don't use auto_now for future
                                        # `bulk_insert` compatibility

    class Meta:
        abstract = True
        unique_together = ('wp', 'id')

    # CUSTOM METHODS #
    def fetch(self):  # pragma: no cover
        """
        Fetch this object from the api and update.
        """
        raise NotImplementedError


##########
# MODELS #
##########

class WPSite(models.Model):
    """
    This describes the WordPress site you're trying to integrate with.

    We only need the url.
    """
    url = models.URLField('URL', unique=True,
        help_text=u'The URL of the WordPress site.')
    name = models.CharField(max_length=100, null=True, blank=True)
    description = models.CharField(max_length=100, null=True, blank=True)
    enable_log = models.BooleanField(default=True,
        help_text='Log requests')

    class Meta:
        verbose_name = u'site'

    def __unicode__(self):
        return self.name or self.url

    def get_absolute_url(self):
        return self.url

    # CUSTOM PROPERTIES #

    @property
    def hook_url(self):
        return reverse('wjordpress:hook_endpoint', kwargs={'pk': self.pk})

    # CUSTOM METHODS #

    def save_from_resource(self, data):
        self.name = data['name']
        self.description = data['description']
        self.save()

    def fetch(self):
        api = WPApi(self.url)
        self.save_from_resource(api.index())
        WPLog.objects.push(self, 'fetch', api.response.text)

    def fetch_all(self):
        api = WPApi(self.url)
        self.save_from_resource(api.index())
        data = api.posts()
        WPPost.objects.get_or_create_from_resource_list(self, data)
        WPLog.objects.push(self, 'fetch_all', api.response.text)


class WPUser(WPObjectModel):
    username = models.CharField(max_length=255)
    name = models.CharField(max_length=255)
    slug = models.SlugField(max_length=255)
    url = models.URLField(null=True, blank=True,
        help_text='The url the author has set as their homepage.')
    avatar = models.URLField(null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    registered = models.DateTimeField(null=True, blank=True)
    # first_name
    # last_name

    # MANAGERS #
    objects = managers.WPManager()

    class Meta(WPObjectModel.Meta):
        verbose_name = u'user'

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return self.url


class WPTag(WPObjectModel):
    name = models.CharField(max_length=255)
    slug = models.SlugField(max_length=255)
    link = models.URLField()

    # MANAGERS #
    objects = managers.WPManager()

    class Meta(WPObjectModel.Meta):
        verbose_name = u'tag'

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return self.link


class WPCategory(WPObjectModel):
    name = models.CharField(max_length=255)
    slug = models.SlugField(max_length=255)
    description = models.TextField(null=True, blank=True)
    link = models.URLField()

    # MANAGERS #
    objects = managers.WPManager()

    class Meta(WPObjectModel.Meta):
        verbose_name = u'category'
        verbose_name_plural = u'categories'

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return self.link


class Image(object):
    """A simple image object."""
    url = None
    width = 0
    height = 0

    def __init__(self, url, width, height, **kwargs):
        self.url = url
        self.width = width
        self.height = height

    def __repr__(self):
        return self.url


class WPPost(WPObjectModel):
    """
    WordPress Posts. This is the main content type.

    more about revisions: http://codex.wordpress.org/Revisions
    """
    title = models.CharField(max_length=255)
    # http://codex.wordpress.org/Post_Status
    status = models.CharField(max_length=20, choices=(
        ('auto-draft', 'Auto-Draft'),
        ('draft', 'Draft'),
        ('future', 'Future'),
        ('inherit', 'Inherit'),
        ('pending', 'Pending'),
        ('private', 'Private'),
        ('publish', 'Published'),
        ('trash', 'Trash'),
    ))
    # http://codex.wordpress.org/Post_Types
    type = models.CharField(max_length=20, choices=(
        ('attachment', 'Attachment'),
        ('nav_menu_item', 'Navication menu'),
        ('page', 'Page'),
        ('post', 'Post'),
        ('revision', 'Revision'),
    ))  # choices? `post`
    content = models.TextField(null=True, blank=True)
    link = models.URLField()
    date = models.DateTimeField()
    modified = models.DateTimeField()
    slug = models.SlugField(max_length=255)
    excerpt = models.TextField(null=True, blank=True)

    author = models.ForeignKey(WPUser, null=True, blank=True)
    categories = models.ManyToManyField(WPCategory, blank=True)
    tags = models.ManyToManyField(WPTag, blank=True)

    parent = models.ForeignKey('self', null=True, blank=True,
        related_name='revisions',
        help_text=u'Revision parent.')
    featured_image = models.ForeignKey('self', null=True, blank=True)
    attachment_meta = JSONField(null=True, blank=True)  # WISHLIST no ext deps

    # MANAGERS #
    objects = managers.WPPostManager()

    class Meta(WPObjectModel.Meta):
        get_latest_by = 'date'
        ordering = ('-date', )
        verbose_name = u'post'

    def __unicode__(self):
        return self.title

    def get_absolute_url(self):
        return self.link

    # `WPObjectModel` METHODS #
    def fetch(self):
        api = WPApi(self.wp.url)
        data = api.posts(self.id)
        WPLog.objects.push(self.wp, 'fetch', api.response.text)
        WPPost.objects.get_or_create_from_resource(self.wp, data)

    # CUSTOM PROPERTIES #
    @property
    def images(self):
        if not self.attachment_meta:
            return None
        sizes = self.attachment_meta['sizes']
        return {
            k: Image(**sizes[k]) for k in sizes
        }


class WPLog(models.Model):
    """Log api communications."""
    wp = models.ForeignKey('WPSite')
    timestamp = models.DateTimeField(auto_now_add=True)
    action = models.CharField(max_length=30)  # TODO use choices so it's easier to filter in admin
    body = models.TextField(default=u'')

    # MANAGERS #
    objects = managers.WPLogManager()

    def __unicode__(self):
        return u'{} {}'.format(self.wp, self.timestamp)

    class Meta:
        ordering = ('-timestamp', )
        verbose_name = 'log'



from django.db import models
from django.utils.timezone import now

import logging


class WPManager(models.Manager):
    def get_or_create_from_resource(self, site, data):
        """
        Get and create or update an instance based on the json.
        """
        field_names = self.model._meta.get_all_field_names()
        obj_data = {k: v for k, v in data.items() if k in field_names}
        obj_data['synced_at'] = now()
        obj, created = self.get_or_create(wp=site, id=data['ID'], defaults=obj_data)
        if not created:
            # obj.__dict__.update does not update foreign keys, but this does
            for k, v in obj_data.items():
                setattr(obj, k, v)
            # Always dirty because 'synced_at' changes
            obj.save()
        return obj, created

    def get_or_create_from_resource_list(self, site, data):
        """
        Get or create multiples from an array json.
        """
        logger = logging.getLogger(__name__)
        for datum in data:
            logger.debug(self.get_or_create_from_resource(site, datum))


class WPPostManager(WPManager):
    def get_or_create_from_resource(self, site, data):
        """
        Get or create an instance based on the json.

        The `WPPost` object has some relations that need to be handled special.
        """
        if 'author' in data:
            from .models import WPUser  # avoid circular imports
            author, __ = WPUser.objects.get_or_create_from_resource(
                site, data['author'])
            data['author'] = author  # XXX mutate original `data` to reference
                                     # something else. I should have made a
                                     # copy, but I'm lazy
        # 'parent' is either `0` or another post json or an int
        parent_data = data.pop('parent')
        if parent_data:
            if isinstance(parent_data, int):
                data['parent'] = self.model.objects.get(wp=site, id=parent_data)
            else:
                # may need to catch DoesNotExist
                data['parent'] = self.model.objects.get(wp=site, id=parent_data['ID'])
        # 'featured_image' isn't always in the resource JSON
        featured_image_data = data.pop('featured_image') if 'featured_image' in data else None

        obj, created = (super(WPPostManager, self)
            .get_or_create_from_resource(site, data))

        if featured_image_data is not None:
            # featured_image_data references the post, so this has to happen
            # after the post is created
            # TODO need to store that this is an image and attachment_meta
            featured_image, __ = self.get_or_create_from_resource(site, featured_image_data)
            # mutate the data to shim `original` into the `sizes` dict
            featured_image_data['attachment_meta']['sizes']['original'] = {
                'file': featured_image_data['attachment_meta']['file'],
                'width': featured_image_data['attachment_meta']['width'],
                'height': featured_image_data['attachment_meta']['height'],
                'mime-type': 'image/jpeg',  # FIXME where do we get this?
                'url': featured_image_data['source'],
            }
            obj.featured_image = featured_image
            obj.save()
        # add many-to-many relations
        if data['terms']:
            category_data = data['terms'].get('category')
            if category_data:
                from .models import WPCategory  # avoid circular imports
                # TODO handle category removal
                for cat_data in category_data:
                    cat, __ = WPCategory.objects.get_or_create_from_resource(
                        site, cat_data)
                    obj.categories.add(cat)
            tags_data = data['terms'].get('post_tag')
            if tags_data:
                from .models import WPTag  # avoid circular imports
                # TODO handle tag removal
                for tag_data in tags_data:
                    tag, __ = WPTag.objects.get_or_create_from_resource(
                        site, tag_data)
                    obj.tags.add(tag)
        return obj, created


class WPLogManager(models.Manager):
    def push(self, wp, action, body=u''):
        """Generic way to create a log entry."""
        if not wp.enable_log:
            return
        self.create(wp=wp, action=action, body=body)



"""
Copyright (c) 2007 Michael Trier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.


https://github.com/django-extensions/django-extensions/blob/b9c996c0818da36d10449dfff7d776bf7dcbbdf2/django_extensions/db/fields/json.py

JSONField automatically serializes most Python terms to JSON data.
Creates a TEXT field with a default value of "{}".  See test_json.py for
more information.

 from django.db import models
 from django_extensions.db.fields import json

 class LOL(models.Model):
     extra = json.JSONField()
"""
from __future__ import absolute_import
import six
from decimal import Decimal
from django.db import models
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

try:
    # Django >= 1.7
    import json
except ImportError:
    # Django <= 1.6 backwards compatibility
    from django.utils import simplejson as json


def dumps(value):
    return DjangoJSONEncoder().encode(value)


def loads(txt):
    value = json.loads(
        txt,
        parse_float=Decimal,
        encoding=settings.DEFAULT_CHARSET
    )
    return value


class JSONDict(dict):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """
    def __repr__(self):
        return dumps(self)


class JSONUnicode(six.text_type):
    """
    As above
    """
    def __repr__(self):
        return dumps(self)


class JSONList(list):
    """
    As above
    """
    def __repr__(self):
        return dumps(self)


class JSONField(six.with_metaclass(models.SubfieldBase, models.TextField)):
    """JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly.  Main thingy must be a dict object."""

    def __init__(self, *args, **kwargs):
        default = kwargs.get('default', None)
        if default is None:
            kwargs['default'] = '{}'
        elif isinstance(default, (list, dict)):
            kwargs['default'] = dumps(default)
        models.TextField.__init__(self, *args, **kwargs)

    def to_python(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if value is None or value == '':
            return {}
        elif isinstance(value, six.string_types):
            res = loads(value)
            if isinstance(res, dict):
                return JSONDict(**res)
            elif isinstance(res, six.string_types):
                return JSONUnicode(res)
            elif isinstance(res, list):
                return JSONList(res)
            return res
        else:
            return value

    def get_db_prep_save(self, value, connection, **kwargs):
        """Convert our JSON object to a string before we save"""
        if value is None and self.null:
            return None
        return super(JSONField, self).get_db_prep_save(dumps(value), connection=connection)

    def south_field_triple(self):
        """Returns a suitable description of this field for South."""
        # We'll just introspect the _actual_ field.
        from south.modelsinspector import introspector
        field_class = "django.db.models.fields.TextField"
        args, kwargs = introspector(self)
        # That's our definition!
        return (field_class, args, kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(JSONField, self).deconstruct()
        if self.default == '{}':
            del kwargs['default']
        return name, path, args, kwargs



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Manually fetch contents.'
    args = '[url]'

    def handle(self, url=None, *args, **options):
        from wjordpress.models import WPSite
        if url is None:
            for site in WPSite.objects.all():
                print u'Fetching {}'.format(site)
                site.fetch_all()
        else:
            try:
                site = WPSite.objects.get(url=url)
                site.fetch_all()
            except WPSite.DoesNotExist:
                print u'Error: Must specify an existing site. Options:'
                for site in WPSite.objects.all():
                    print u'  * {}'.format(site.url)



import logging

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from . import models
from .api import UserCannotReadException


logger = logging.getLogger(__name__)


class HookPressEndpoint(View):

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super(HookPressEndpoint, self).dispatch(*args, **kwargs)

    def post(self, request, pk, **kwargs):
        hook_used = request.POST.get('hook')
        site = models.WPSite.objects.get(pk=pk)
        logger.info(u'Hook Triggered: {}'.format(hook_used), extra={
            'request': request,
        })
        # TODO clean up request.POST body so it's more useful
        models.WPLog.objects.push(site, 'hook', request.POST)
        if hook_used == 'save_post':
            try:
                post = models.WPPost.objects.get(
                    wp=site,
                    id=request.POST.get('ID'),
                )
            except models.WPPost.DoesNotExist:
                # need to create
                post = models.WPPost(wp=site, id=request.POST.get('ID'))
            try:
                post.fetch()
            except UserCannotReadException as e:
                # downgrade exception to INFO
                logger.info(e)
        return HttpResponse('OK')



from django.contrib import admin
from django.core.urlresolvers import NoReverseMatch

from . import models


class WPSiteAdmin(admin.ModelAdmin):
    list_display = ('name', 'url', 'hook')
    readonly_fields = ('name', 'description')

    def save_model(self, request, obj, form, change):
        # TODO do this sync async (give celery another shot?)
        obj.save()
        obj.fetch_all()

    # CUSTOM METHODS #

    def hook(self, obj):
        """
        This is where an admin can find what url to point the webhook to.

        Doing it as an absolute url lets us cheat and make the browser figure
        out the host for us.

        Requires HookPress: http://wordpress.org/plugins/hookpress/
        """
        try:
            return (u'<a href="{}" title="Add a save_post hook with the ID">'
                'Webhook</a>'.format(obj.hook_url))
        except NoReverseMatch:
            return ''
    hook.allow_tags = True
admin.site.register(models.WPSite, WPSiteAdmin)


class WPUserAdmin(admin.ModelAdmin):
    readonly_fields = ('synced_at', )
admin.site.register(models.WPUser, WPUserAdmin)


class WPCategoryAdmin(admin.ModelAdmin):
    readonly_fields = ('synced_at', )
admin.site.register(models.WPCategory, WPCategoryAdmin)


class WPTagAdmin(admin.ModelAdmin):
    readonly_fields = ('synced_at', )
admin.site.register(models.WPTag, WPTagAdmin)


class WPPostAdmin(admin.ModelAdmin):
    list_display = ('title', 'date', 'type', 'status', )
    list_filter = ('type', 'status', )
    readonly_fields = ('synced_at', )
admin.site.register(models.WPPost, WPPostAdmin)


class WPLogAdmin(admin.ModelAdmin):
    list_display = ('timestamp', 'wp', 'action', )
    list_filter = ('wp', 'action', )
    readonly_fields = ('wp', 'timestamp', 'action', 'body', )
admin.site.register(models.WPLog, WPLogAdmin)



from django import template

from ..models import WPSite, WPPost


register = template.Library()


@register.inclusion_tag('wjordpress/widget.html')
def wjidget(wp_site_name, limit=10):
    """
    This renders a widget you can use to quickly show wordpress content.

    Pass in the name of the wordpress site.
    TODO the name of the site is a brittle way to reference, but it seems more
    human friendly than the url.
    """
    wp_site = WPSite.objects.get(name=wp_site_name)
    return {
        'object_list': WPPost.objects.filter(
            wp=wp_site,
            status='publish',  # XXX magic constant
            type='post',  # XXX magic constant
        )[:limit],
    }



from setuptools import setup

setup(
    name='wjordpress',
    # when bumping versions, also update __init__, sphinx config, and changelog
    version='0.2.2',
    author='Chris Chang',
    author_email='cchang@texastribune.org',
    url='https://github.com/texastribune/wjordpress',
    packages=['wjordpress'],
    include_package_data=True,  # automatically include things from MANIFEST.in
    license='Apache License, Version 2.0',
    description='Django integration with WordPress through the json-rest-api plugin',
    long_description=open('README.rst').read(),
    install_requires=[
        'six>=1.0.0',  # works in tox
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)
